@@ -25,13 +25,20 @@
 #include "grd-session-vnc.h"
 
 #include <gio/gio.h>
-#include <gst/gst.h>
 #include <linux/input.h>
 #include <rfb/rfb.h>
 
+#include "grd-context.h"
+#include "grd-prompt.h"
+#include "grd-settings.h"
 #include "grd-stream.h"
 #include "grd-vnc-server.h"
-#include "grd-vnc-sink.h"
+#include "grd-vnc-pipewire-stream.h"
+
+/* BGRx */
+#define BGRX_BITS_PER_SAMPLE 8
+#define BGRX_SAMPLES_PER_PIXEL 3
+#define BGRX_BYTES_PER_PIXEL 4
 
 struct _GrdSessionVnc
 {
@@ -42,9 +49,18 @@ struct _GrdSessionVnc
   rfbScreenInfoPtr rfb_screen;
   rfbClientPtr rfb_client;
 
-  GstElement *pipeline;
-  GstElement *vnc_sink;
-  GstElement *pipewire_src;
+  gboolean pending_framebuffer_resize;
+  int pending_framebuffer_width;
+  int pending_framebuffer_height;
+
+  guint close_session_idle_id;
+
+  GrdVncAuthMethod auth_method;
+
+  GrdPrompt *prompt;
+  GCancellable *prompt_cancellable;
+
+  GrdVncPipeWireStream *pipewire_stream;
 
   int prev_x;
   int prev_y;
@@ -57,58 +73,89 @@ G_DEFINE_TYPE (GrdSessionVnc, grd_session_vnc, GRD_TYPE_SESSION);
 static void
 grd_session_vnc_detach_source (GrdSessionVnc *session_vnc);
 
-void
-grd_session_vnc_resize_framebuffer (GrdSessionVnc *session_vnc,
-                                    int            width,
-                                    int            height)
+static gboolean
+close_session_idle (gpointer user_data);
+
+static rfbBool
+check_rfb_password (rfbClientPtr  rfb_client,
+                    const char   *response_encrypted,
+                    int           len);
+
+static void
+swap_uint8 (uint8_t *a,
+            uint8_t *b)
+{
+  uint8_t tmp;
+
+  tmp = *a;
+  *a = *b;
+  *b = tmp;
+}
+
+static void
+resize_vnc_framebuffer (GrdSessionVnc *session_vnc,
+                        int            width,
+                        int            height)
 {
   rfbScreenInfoPtr rfb_screen = session_vnc->rfb_screen;
   uint8_t *framebuffer;
 
+  if (!session_vnc->rfb_client->useNewFBSize)
+    g_warning ("Client does not support NewFBSize");
+
+  g_free (rfb_screen->frameBuffer);
+  framebuffer = g_malloc0 (width * height * BGRX_BYTES_PER_PIXEL);
+  rfbNewFramebuffer (rfb_screen,
+                     (char *) framebuffer,
+                     width, height,
+                     BGRX_BITS_PER_SAMPLE,
+                     BGRX_SAMPLES_PER_PIXEL,
+                     BGRX_BYTES_PER_PIXEL);
+
+  /*
+   * Our format is hard coded to BGRX but LibVNCServer asusumes it's RGBX;
+   * lets override that.
+   */
+  swap_uint8 (&session_vnc->rfb_screen->serverFormat.redShift,
+              &session_vnc->rfb_screen->serverFormat.blueShift);
+}
+
+void
+grd_session_vnc_queue_resize_framebuffer (GrdSessionVnc *session_vnc,
+                                          int            width,
+                                          int            height)
+{
+  rfbScreenInfoPtr rfb_screen = session_vnc->rfb_screen;
+
   if (rfb_screen->width == width && rfb_screen->height == height)
     return;
 
-  g_free (session_vnc->rfb_screen->frameBuffer);
-  framebuffer = g_malloc0 (width * height * 4);
-  rfbNewFramebuffer (session_vnc->rfb_screen,
-                     (char *) framebuffer,
-                     width, height,
-                     8, 3, 4);
-}
+  if (session_vnc->rfb_client->preferredEncoding == -1)
+    {
+      session_vnc->pending_framebuffer_resize = TRUE;
+      session_vnc->pending_framebuffer_width = width;
+      session_vnc->pending_framebuffer_height = height;
+      return;
+    }
 
-static size_t
-grd_session_vnc_get_framebuffer_size (GrdSessionVnc *session_vnc)
-{
-  return session_vnc->rfb_screen->width * session_vnc->rfb_screen->height * 4;
+  resize_vnc_framebuffer (session_vnc, width, height);
 }
 
 void
 grd_session_vnc_draw_buffer (GrdSessionVnc *session_vnc,
-                             GstBuffer     *buffer)
+                             void          *data)
 {
-  GstMapInfo map_info = { 0 };
-  size_t framebuffer_size = grd_session_vnc_get_framebuffer_size (session_vnc);
+  size_t size;
 
-  if (gst_buffer_get_size (buffer) != framebuffer_size)
-    {
-      g_warning ("VNC sink buffer size %lu differs from VNC framebuffer size\n",
-                 gst_buffer_get_size (buffer));
-      return;
-    }
+  if (session_vnc->pending_framebuffer_resize)
+    return;
 
-  if (!gst_buffer_map (buffer, &map_info, GST_MAP_READ))
-    {
-      g_warning ("Failed to map VNC sink buffer\n");
-      return;
-    }
-
-  g_assert (map_info.size == framebuffer_size);
+  size = (session_vnc->rfb_screen->height *
+          grd_session_vnc_get_framebuffer_stride (session_vnc));
 
   memcpy (session_vnc->rfb_screen->frameBuffer,
-          map_info.data,
-          framebuffer_size);
-
-  gst_buffer_unmap (buffer, &map_info);
+          data,
+          size);
 
   rfbMarkRectAsModified (session_vnc->rfb_screen, 0, 0,
                          session_vnc->rfb_screen->width,
@@ -121,31 +168,99 @@ handle_client_gone (rfbClientPtr rfb_client)
 {
   GrdSessionVnc *session_vnc = rfb_client->screen->screenData;
 
-  /* Avoid re-entry when stopping. */
-  if (!session_vnc->rfb_screen)
-    return;
+  g_debug ("VNC client gone");
 
-  grd_session_stop (GRD_SESSION (session_vnc));
+  session_vnc->close_session_idle_id =
+    g_idle_add (close_session_idle, session_vnc);
 }
 
-static gboolean
-is_session_ready (GrdSessionVnc *session_vnc)
+static void
+prompt_response_callback (GObject      *source_object,
+                          GAsyncResult *async_result,
+                          gpointer      user_data)
 {
-  return session_vnc->pipeline != NULL;
+  GrdSessionVnc *session_vnc = GRD_SESSION_VNC (user_data);
+  g_autoptr(GError) error = NULL;
+  GrdPromptResponse response;
+
+  if (!grd_prompt_query_finish (session_vnc->prompt,
+                                async_result,
+                                &response,
+                                &error))
+    {
+      if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+        {
+          g_warning ("Failed to query user about session: %s", error->message);
+          rfbRefuseOnHoldClient (session_vnc->rfb_client);
+          return;
+        }
+      else
+        {
+          return;
+        }
+    }
+
+  switch (response)
+    {
+    case GRD_PROMPT_RESPONSE_ACCEPT:
+      grd_session_start (GRD_SESSION (session_vnc));
+      rfbStartOnHoldClient (session_vnc->rfb_client);
+      return;
+    case GRD_PROMPT_RESPONSE_REFUSE:
+      rfbRefuseOnHoldClient (session_vnc->rfb_client);
+      return;
+    }
+
+  g_assert_not_reached ();
 }
 
 static enum rfbNewClientAction
 handle_new_client (rfbClientPtr rfb_client)
 {
   GrdSessionVnc *session_vnc = GRD_SESSION_VNC (rfb_client->screen->screenData);
+  GrdContext *context = grd_session_get_context (GRD_SESSION (session_vnc));
+  GrdSettings *settings = grd_context_get_settings (context);
+
+  g_debug ("New VNC client");
+
+  session_vnc->auth_method = grd_settings_get_vnc_auth_method (settings);
 
   session_vnc->rfb_client = rfb_client;
   rfb_client->clientGoneHook = handle_client_gone;
 
-  if (!is_session_ready (session_vnc))
-    grd_session_vnc_detach_source (session_vnc);
+  switch (session_vnc->auth_method)
+    {
+    case GRD_VNC_AUTH_METHOD_PROMPT:
+      session_vnc->prompt = g_object_new (GRD_TYPE_PROMPT, NULL);
+      session_vnc->prompt_cancellable = g_cancellable_new ();
+      grd_prompt_query_async (session_vnc->prompt,
+                              rfb_client->host,
+                              session_vnc->prompt_cancellable,
+                              prompt_response_callback,
+                              session_vnc);
+      grd_session_vnc_detach_source (session_vnc);
+      return RFB_CLIENT_ON_HOLD;
+    case GRD_VNC_AUTH_METHOD_PASSWORD:
+      session_vnc->rfb_screen->passwordCheck = check_rfb_password;
+      /*
+       * authPasswdData needs to be non NULL in libvncserver to trigger
+       * password authentication.
+       */
+      session_vnc->rfb_screen->authPasswdData = (gpointer) 1;
 
-  return RFB_CLIENT_ACCEPT;
+      return RFB_CLIENT_ACCEPT;
+    }
+
+  g_assert_not_reached ();
+}
+
+static gboolean
+is_view_only (GrdSessionVnc *session_vnc)
+{
+  GrdContext *context = grd_session_get_context (GRD_SESSION (session_vnc));
+  GrdSettings *settings = grd_context_get_settings (context);
+
+  return grd_settings_get_vnc_view_only (settings);
 }
 
 static void
@@ -156,6 +271,9 @@ handle_key_event (rfbBool      down,
   GrdSessionVnc *session_vnc = GRD_SESSION_VNC (rfb_client->screen->screenData);
   GrdSession *session = GRD_SESSION (session_vnc);
 
+  if (is_view_only (session_vnc))
+    return;
+
   if (down)
     {
       if (g_hash_table_add (session_vnc->pressed_keys,
@@ -210,22 +328,22 @@ grd_session_vnc_notify_axis (GrdSessionVnc *session_vnc,
 
   switch (button_mask_bit_index)
     {
-    case 4:
+    case 3:
       axis = GRD_POINTER_AXIS_VERTICAL;
       steps = -1;
       break;
 
-    case 5:
+    case 4:
       axis = GRD_POINTER_AXIS_VERTICAL;
       steps = 1;
       break;
 
-    case 6:
+    case 5:
       axis = GRD_POINTER_AXIS_HORIZONTAL;
       steps = -1;
       break;
 
-    case 7:
+    case 6:
       axis = GRD_POINTER_AXIS_HORIZONTAL;
       steps = 1;
       break;
@@ -246,6 +364,9 @@ handle_pointer_event (int          button_mask,
   GrdSessionVnc *session_vnc = rfb_client->screen->screenData;
   GrdSession *session = GRD_SESSION (session_vnc);
 
+  if (is_view_only (session_vnc))
+    return;
+
   if (x != session_vnc->prev_x || y != session_vnc->prev_y)
     {
       grd_session_notify_pointer_motion_absolute (session, x, y);
@@ -258,15 +379,15 @@ handle_pointer_event (int          button_mask,
     {
       unsigned int i;
       int buttons[] = {
-        BTN_LEFT,   /* 1 */
-        BTN_MIDDLE, /* 2 */
-        BTN_RIGHT,  /* 3 */
-        0,          /* 4 - vertical scroll: up */
-        0,          /* 5 - vertical scroll: down */
-        0,          /* 6 - horizontal scroll: left */
-        0,          /* 7 - horizontal scroll: right */
-        BTN_SIDE,   /* 8 */
-        BTN_EXTRA,  /* 9 */
+        BTN_LEFT,   /* 0 */
+        BTN_MIDDLE, /* 1 */
+        BTN_RIGHT,  /* 2 */
+        0,          /* 3 - vertical scroll: up */
+        0,          /* 4 - vertical scroll: down */
+        0,          /* 5 - horizontal scroll: left */
+        0,          /* 6 - horizontal scroll: right */
+        BTN_SIDE,   /* 7 */
+        BTN_EXTRA,  /* 8 */
       };
 
       for (i = 0; i < G_N_ELEMENTS (buttons); i++)
@@ -296,6 +417,60 @@ handle_pointer_event (int          button_mask,
   rfbDefaultPtrAddEvent (button_mask, x, y, rfb_client);
 }
 
+static rfbBool
+check_rfb_password (rfbClientPtr  rfb_client,
+                    const char   *response_encrypted,
+                    int           len)
+{
+  GrdSessionVnc *session_vnc = rfb_client->screen->screenData;
+  GrdContext *context = grd_session_get_context (GRD_SESSION (session_vnc));
+  GrdSettings *settings = grd_context_get_settings (context);
+  g_autofree char *password = NULL;
+  g_autoptr(GError) error = NULL;
+  uint8_t challenge_encrypted[CHALLENGESIZE];
+
+  switch (session_vnc->auth_method)
+    {
+    case GRD_VNC_AUTH_METHOD_PROMPT:
+      return TRUE;
+    case GRD_VNC_AUTH_METHOD_PASSWORD:
+      break;
+    }
+
+  password = grd_settings_get_vnc_password (settings, &error);
+  if (!password)
+    {
+      g_warning ("Couldn't retrieve VNC password: %s", error->message);
+      return FALSE;
+    }
+
+  if (strlen (password) == 0)
+    {
+      g_warning ("VNC password was empty, denying");
+      return FALSE;
+    }
+
+  memcpy(challenge_encrypted, rfb_client->authChallenge, CHALLENGESIZE);
+  rfbEncryptBytes(challenge_encrypted, password);
+
+  if (memcmp (challenge_encrypted, response_encrypted, len) == 0)
+    {
+      grd_session_start (GRD_SESSION (session_vnc));
+      grd_session_vnc_detach_source (session_vnc);
+      return TRUE;
+    }
+  else
+    {
+      return FALSE;
+    }
+}
+
+int
+grd_session_vnc_get_framebuffer_stride (GrdSessionVnc *session_vnc)
+{
+  return session_vnc->rfb_screen->paddedWidthInBytes;
+}
+
 static void
 init_vnc_session (GrdSessionVnc *session_vnc)
 {
@@ -318,6 +493,9 @@ init_vnc_session (GrdSessionVnc *session_vnc)
   rfb_screen->newClientHook = handle_new_client;
   rfb_screen->screenData = session_vnc;
 
+  /* Amount of milliseconds to wait before sending an update */
+  rfb_screen->deferUpdateTime = 0;
+
   rfb_screen->kbdAddEvent = handle_key_event;
   rfb_screen->kbdReleaseAllKeys = handle_release_all_keys;
   rfb_screen->ptrAddEvent = handle_pointer_event;
@@ -325,10 +503,10 @@ init_vnc_session (GrdSessionVnc *session_vnc)
   rfb_screen->frameBuffer = g_malloc0 (screen_width * screen_height * 4);
   memset (rfb_screen->frameBuffer, 0x1f, screen_width * screen_height * 4);
 
+  session_vnc->rfb_screen = rfb_screen;
+
   rfbInitServer (rfb_screen);
   rfbProcessEvents (rfb_screen, 0);
-
-  session_vnc->rfb_screen = rfb_screen;
 }
 
 static gboolean
@@ -341,14 +519,26 @@ handle_socket_data (GSocket *socket,
   if (condition & G_IO_IN)
     {
       if (rfbIsActive (session_vnc->rfb_screen))
-        rfbProcessEvents (session_vnc->rfb_screen, 0);
+        {
+          rfbProcessEvents (session_vnc->rfb_screen, 0);
+
+          if (session_vnc->pending_framebuffer_resize &&
+              session_vnc->rfb_client->preferredEncoding != -1)
+            {
+              resize_vnc_framebuffer (session_vnc,
+                                      session_vnc->pending_framebuffer_width,
+                                      session_vnc->pending_framebuffer_height);
+              session_vnc->pending_framebuffer_resize = FALSE;
+            }
+        }
     }
   else
     {
-      g_warning ("Unhandled socket condition %d\n", condition);
+      g_debug ("Unhandled socket condition %d\n", condition);
+      return G_SOURCE_REMOVE;
     }
 
-  return TRUE;
+  return G_SOURCE_CONTINUE;
 }
 
 static void
@@ -410,12 +600,16 @@ grd_session_vnc_stop (GrdSession *session)
 {
   GrdSessionVnc *session_vnc = GRD_SESSION_VNC (session);
 
-  if (session_vnc->pipeline)
+  g_debug ("Stopping VNC session");
+
+  if (session_vnc->close_session_idle_id)
     {
-      gst_element_set_state (session_vnc->pipeline, GST_STATE_NULL);
-      g_clear_pointer (&session_vnc->pipeline, gst_object_unref);
+      g_source_remove (session_vnc->close_session_idle_id);
+      session_vnc->close_session_idle_id = 0;
     }
 
+  g_clear_object (&session_vnc->pipewire_stream);
+
   grd_session_vnc_detach_source (session_vnc);
 
   g_clear_object (&session_vnc->connection);
@@ -423,54 +617,49 @@ grd_session_vnc_stop (GrdSession *session)
   g_clear_pointer (&session_vnc->rfb_screen, (GDestroyNotify) rfbScreenCleanup);
 }
 
+static gboolean
+close_session_idle (gpointer user_data)
+{
+  GrdSessionVnc *session_vnc = GRD_SESSION_VNC (user_data);
+
+  grd_session_stop (GRD_SESSION (session_vnc));
+
+  session_vnc->close_session_idle_id = 0;
+
+  return G_SOURCE_REMOVE;
+}
+
+static void
+on_pipwire_stream_closed (GrdVncPipeWireStream *stream,
+                          GrdSessionVnc        *session_vnc)
+{
+  g_warning ("PipeWire stream closed, closing client");
+
+  session_vnc->close_session_idle_id =
+    g_idle_add (close_session_idle, session_vnc);
+}
+
 static void
 grd_session_vnc_stream_ready (GrdSession *session,
                               GrdStream  *stream)
 {
   GrdSessionVnc *session_vnc = GRD_SESSION_VNC (session);
   uint32_t pipewire_node_id;
-  g_autofree char *pipeline_str = NULL;
-  GError *error = NULL;
-  g_autoptr(GstElement) pipeline = NULL;
-  g_autoptr(GstElement) vnc_sink = NULL;
-  g_autoptr(GstPad) sink_pad = NULL;
-  g_autoptr(GstPad) src_pad = NULL;
+  g_autoptr (GError) error = NULL;
 
   pipewire_node_id = grd_stream_get_pipewire_node_id (stream);
-  pipeline_str =
-    g_strdup_printf ("pipewiresrc name=pipewiresrc path=%d ! videoconvert",
-                     pipewire_node_id);
-
-  pipeline = gst_parse_launch (pipeline_str, &error);
-  if (!pipeline)
+  session_vnc->pipewire_stream = grd_vnc_pipewire_stream_new (session_vnc,
+                                                              pipewire_node_id,
+                                                              &error);
+  if (!session_vnc->pipewire_stream)
     {
-      g_warning ("Failed to start VNC pipeline: %s\n", error->message);
+      g_warning ("Failed to establish PipeWire stream: %s", error->message);
       return;
     }
 
-  vnc_sink = grd_vnc_sink_new (session_vnc);
-  if (!vnc_sink)
-    {
-      g_warning ("Failed to create VNC sink\n");
-      return;
-    }
-  gst_bin_add (GST_BIN (pipeline), vnc_sink);
-
-  sink_pad = gst_element_get_static_pad (vnc_sink, "sink");
-  src_pad = gst_bin_find_unlinked_pad (GST_BIN (pipeline), GST_PAD_SRC);
-
-  if (gst_pad_link (src_pad, sink_pad) != GST_PAD_LINK_OK)
-    {
-      g_warning ("Failed to link VNC sink to pipeline\n");
-      return;
-    }
-
-  gst_element_set_state (pipeline, GST_STATE_PLAYING);
-
-  session_vnc->pipewire_src = gst_bin_get_by_name (GST_BIN (pipeline),
-                                                   "pipewiresrc");
-  session_vnc->vnc_sink = g_steal_pointer (&vnc_sink);
-  session_vnc->pipeline = g_steal_pointer (&pipeline);
+  g_signal_connect (session_vnc->pipewire_stream, "closed",
+                    G_CALLBACK (on_pipwire_stream_closed),
+                    session_vnc);
 
   if (!session_vnc->source)
     grd_session_vnc_attach_source (session_vnc);
